﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using FTDI.D2XX.Device;
using System.Threading.Tasks;
using System.Diagnostics;
namespace DeviceWatcher
{
    public class MPSSEHAL : IDisposable
    {
        private FTManager ftManager;
        private FTD2xxDevice device = null;

        const byte mpsseIN = 0x10;
        const byte mpsseOut = 0x20;


        const uint CLOCK_6MHZ = 6000000;
        const uint CLOCK_30MHZ = 30000000;

        private enum CMD : byte
        {

             ECHO_COMMAND_ONCE = 0,
             ECHO_COMMAND_CONTINUOUSLY = 1,
             ECHO_CMD_1 = 0xAA,
             ECHO_CMD_2 = 0xAB,
             BAD_COMMAND_RESPONSE = 0xFA,
             CMD_NOT_ECHOED = 0,
             CMD_ECHOED = 1,            
             SET_LOW_BYTE_DATA_BITS = 0x80,
             GET_LOW_BYTE_DATA_BITS = 0x81,
             SET_HIGH_BYTE_DATA_BITS = 0x82,
             GET_HIGH_BYTE_DATA_BITS = 0x83,
             MID_TURN_ON_LOOPBACK = 0x84,
             MID_TURN_OFF_LOOPBACK = 0x85,
             SET_CLOCK_FREQUENCY = 0x86,     
             DISABLE_CLOCK_DIVIDE = 0x8A,
             ENABLE_CLOCK_DIVIDE = 0x8B
    }

    const byte GPIO_LOW_BYTE_DATA_BITS = 0x2B;  
    const byte GPIO_HIGH_BYTE_DATA_BITS = 0x0F;

    const byte LOOPBACK_FALSE                       = 0;
        const byte LOOPBACK_TRUE                        = 1;
        const byte CMD_MID_TURN_ON_LOOPBACK             = 0x84;
        const byte CMD_MID_TURN_OFF_LOOPBACK            = 0x85;

        const byte CS_MASK  = 0x08;
        const byte PD_MASK  = 0x20;




        /*MPSSE Data Commands - byte mode - MSB first */
        const byte MPSSE_CMD_DATA_OUT_BYTES_POS_EDGE        = 0x10;
        const byte MPSSE_CMD_DATA_OUT_BYTES_NEG_EDGE        = 0x11;
        const byte MPSSE_CMD_DATA_IN_BYTES_POS_EDGE         = 0x20;
        const byte MPSSE_CMD_DATA_IN_BYTES_NEG_EDGE         = 0x24;
        const byte MPSSE_CMD_DATA_BYTES_IN_POS_OUT_NEG_EDGE = 0x31;
        const byte MPSSE_CMD_DATA_BYTES_IN_NEG_OUT_POS_EDGE = 0x34;
        public MPSSEHAL()
        {
            ftManager = new FTManager();
            if (device != null)
                device.Close();            
        }

        async public Task Configure (FTD2xxDevice inDevice)
        {
            device = inDevice;
            if (device == null)
            {
                return;
            }

            if (device.IsOpen)
            {
                byte[] cmd = { 0xaa };
                byte[] rd = new byte[2];
                byte data;

                try
                {
                    await device.SetTimeouts((ushort)5000, (ushort)5000);
                    await device.SetLatencyTimerAsync(16);
                    await device.SetFlowControlAsync((ushort) FLOW_CONTROL.RTS_CTS, 0x00, 0x00);
                    //await device.SetFlowControlAsync((ushort) FLOW_CONTROL.NONE, 0x00, 0x00);
                    await device.SetBitModeAsync((byte) BIT_MODE.RESET, 0x00);
                    await device.SetBitModeAsync((byte) BIT_MODE.MPSSE, 0x02);


                    await Task.Delay(50);
                    uint num;
                    cmd[0] = 0xAA;
                    num = await device.WriteAsync(cmd, 1);
                    uint c;
                    while ((c = await device.GetQueueStatus()) == 0)                    
                        await Task.Delay(50);
                    
                    num = await device.ReadAsync(rd, c);
                    if ( num >= 1 && rd[0] == 0xFA )                  
                        Debug.WriteLine("OK!");
                    else
                        Debug.WriteLine("Not OK!");                    

                    await SetClock();

                    await SetGPIO();
                    await PDLow();
                    await Task.Delay(50);                   
                    await PDHigh();
                    
                }
                catch ( Exception ex)
                {
                    Debug.WriteLine("MPSSEHAL: exception, {0}", ex.Message);
                }

            }
           
            return;
        }


        //async public Task ReadWrite( byte[] inBuff,  byte[] outBuff  )
        async public Task<byte> ReadWrite(byte inBuff)
        {
            byte[] rdBuff = new byte[1];
            byte[] wrBuff = new byte[1]{ inBuff };

            rdBuff = await ReadWrite(wrBuff);
            return rdBuff[0];
        }

        async public Task<byte[]> ReadWrite(byte[] data)
        {
            uint num = (uint)data.Length + 3 ;
            uint avail;
            byte lengthLSB = (byte)(data.Length - 1 );
            byte lengthMSB = 0; // (byte)((data.Length -1) >>8);
            byte[] wrBuff = new byte[data.Length + 3 +1];

            if (num > 255)
                throw (new Exception("exceeds array size"));

            wrBuff[0] = MPSSE_CMD_DATA_BYTES_IN_POS_OUT_NEG_EDGE;
            wrBuff[1] = lengthLSB;
            wrBuff[2] = lengthMSB;
            data.CopyTo(wrBuff, 3);
            wrBuff[num] = 0x87;
            
            byte[] rdBuff;
            //num = (uint)wrBuff.Length;
            await device.WriteAsync(wrBuff, num+1);

            while ((avail = await device.GetQueueStatus()) < data.Length)
                await Task.Delay(1);
            rdBuff = new byte[data.Length];
            await device.ReadAsync(rdBuff, (uint)data.Length);           

            return rdBuff;
        }


        async public Task Write(byte[] data)
        {
            uint num = (uint)data.Length + 3;
            uint avail;
            byte lengthLSB = (byte)(data.Length - 1);
            byte lengthMSB = 0; // (byte)((data.Length -1) >>8);
            byte[] wrBuff = new byte[data.Length + 3 ];

            if (num > 255)
                throw (new Exception("exceeds array size"));

            wrBuff[0] = MPSSE_CMD_DATA_OUT_BYTES_NEG_EDGE;
            wrBuff[1] = lengthLSB;
            wrBuff[2] = lengthMSB;
            data.CopyTo(wrBuff, 3);
 
            await device.WriteAsync(wrBuff, num);

            return;
        }

 


        async public Task CSLow()
        {
            byte data = 0xF6; //set CS low
            /*set the GPIO*/
            byte[] cmd = new byte[] { (byte)CMD.SET_LOW_BYTE_DATA_BITS, data, GPIO_LOW_BYTE_DATA_BITS };
            uint num = await device.WriteAsync(cmd, (uint)cmd.Length);
        }
        async public Task CSHigh()
        {
            byte data = 0xFE; //set CS 
            /*set the GPIO*/
            byte[] cmd = new byte[] { (byte)CMD.SET_LOW_BYTE_DATA_BITS, data, GPIO_LOW_BYTE_DATA_BITS };
            uint num = await device.WriteAsync(cmd, (uint)cmd.Length);
        }


        async public Task PDLow()
        {
            byte data;
            const byte mask = 0xdf;
            /// read GPIO            
            byte[] cmd = new byte[] { (byte)CMD.GET_LOW_BYTE_DATA_BITS, 0x87 };

            uint num = await device.WriteAsync(cmd, (uint)cmd.Length);

            while ((num = await device.GetQueueStatus()) == 0)
                await Task.Delay(5);

            byte[] rdBuf = new byte[1];

            num = await device.ReadAsync(rdBuf, 1);

            data = rdBuf[0];
            data &= mask; //set CS low
            /*set the GPIO*/
            cmd = new byte[] { (byte)CMD.SET_LOW_BYTE_DATA_BITS, data, GPIO_LOW_BYTE_DATA_BITS };

            num = await device.WriteAsync(cmd, (uint)cmd.Length);
        }
        async public Task PDHigh()
        {
            byte data;
            const byte mask = 0x20;
            /// read GPIO            
            byte[] cmd = new byte[] { (byte)CMD.GET_LOW_BYTE_DATA_BITS };

            uint num = await device.WriteAsync(cmd, (uint)cmd.Length);

            while ((num = await device.GetQueueStatus()) == 0)
                await Task.Delay(5);

            byte[] rdBuf = new byte[1];

            num = await device.ReadAsync(rdBuf, 1);
            //Debug.WriteLine("read GPIO LOW: value=0x{0:x}\n", rdBuf[0]);
            data = rdBuf[0];
            data |= mask; //set CS low
            /*set the GPIO*/
            cmd = new byte[] { (byte)CMD.SET_LOW_BYTE_DATA_BITS, data, GPIO_LOW_BYTE_DATA_BITS };

            num = await device.WriteAsync(cmd, (uint)cmd.Length);
        }

        async public Task AllLow()
        {
            byte data;
           
            data = 0x00; //set All low
            /*set the GPIO*/
            byte[] cmd = new byte[] { (byte)CMD.SET_LOW_BYTE_DATA_BITS, data, 0xFF };
            uint num = await device.WriteAsync(cmd, (uint)cmd.Length);
        }
        async public Task AllHigh()
        {
            byte data;
           
            data = 0x28; //set All High
            
            byte[] cmd = new byte[] { (byte)CMD.SET_LOW_BYTE_DATA_BITS, data, 0xFF };
            uint num = await device.WriteAsync(cmd, (uint)cmd.Length);
        }
        async public Task SetClock( uint clock = CLOCK_6MHZ + 1 )
        {
            //uint clock = CLOCK_6MHZ+1;
            short divider;
            uint num;

            byte[] cmd;
            if (clock <= CLOCK_6MHZ)
            {
                cmd =new byte[] { (byte)CMD.ENABLE_CLOCK_DIVIDE };
                Debug.WriteLine("ENABLE_CLOCK_DIVIDE");
                num = await device.WriteAsync(cmd, (uint)cmd.Length);
                divider = (short) ((CLOCK_6MHZ / clock) - 1);
            }
            else
            {
                Debug.WriteLine("DISABLE_CLOCK_DIVIDE");
                cmd = new byte[]{ (byte)CMD.DISABLE_CLOCK_DIVIDE};
                num = await device.WriteAsync(cmd, (uint)cmd.Length);                                
                divider = (short)((CLOCK_30MHZ / clock) - 1);
            }
                    
            /*calculate valueH and ValueL*/
            byte valueL = (byte)divider;
	        byte valueH = (byte)(divider>>8);
	        /*set the clock*/
	        cmd = new byte[]{ (byte)CMD.SET_CLOCK_FREQUENCY, valueL, valueH };
            Debug.WriteLine("set clock: valueL=0x{0:x} valueH=0x{1:x} \n", valueL, valueH);
            var wrote = await device.WriteAsync(cmd, (uint)cmd.Length);
            if (wrote == 3)
                Debug.WriteLine("Clock set.");
            else   
                Debug.WriteLine("Command to set clock failed!");
            return;
        }



        async private Task SetGPIO()
        {
            
            /*set the GPIO*/
            byte[] cmd = new byte[] { (byte)CMD.SET_LOW_BYTE_DATA_BITS, 0x28, GPIO_LOW_BYTE_DATA_BITS };
            uint num  = await device.WriteAsync(cmd, (uint)cmd.Length);
        }

        async private Task TestRead()
        {

            byte[] data = await ReadWrite(new byte[] { 0x00, 0xFF });
            
        }

        #region IDisposable Support
        private bool disposedValue = false; // To detect redundant calls

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    // TODO: dispose managed state (managed objects).
                }

                // TODO: free unmanaged resources (unmanaged objects) and override a finalizer below.
                // TODO: set large fields to null.

                disposedValue = true;
            }
        }

        // TODO: override a finalizer only if Dispose(bool disposing) above has code to free unmanaged resources.
        // ~MPSSEHAL() {
        //   // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
        //   Dispose(false);
        // }

        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
            Dispose(true);
            // TODO: uncomment the following line if the finalizer is overridden above.
            // GC.SuppressFinalize(this);
        }
        #endregion
    }
}